/*
 * $Id: sap.c,v 1.7 2010/02/21 17:21:34 simimeie Exp $
 *
 * This implements a Mini-SAP/SDP Service.
 * It is used by getstream, but can also be compiled with -DSTANDALONE
 * in which case you get a small app for announcing arbitrary SAP/SDP
 * sessions.
 *
 * RFC 2974 (Session Announcement Protocol)
 * http://www.ietf.org/rfc/rfc2974.txt
 * RFC 2327 (SDP: Session Description Protocol)
 * http://www.ietf.org/rfc/rfc2327.txt
 *
 */

#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdint.h>
#include <event.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef STANDALONE

#include <getopt.h>

/* These structs contain only the things necessary for us in standalone mode. */
struct channel {
	unsigned int		sid;
	char			*name;
};

struct stream {
	struct stream *		next;
	struct channel *	channel;
	int			nuccs; /* Number of UniCast Clients; -1 = multicast */
	char			ipaddr[128];
	int			port;
	int			ttl;
	int			isvideo;
};

struct stream * streams = NULL; /* List of streams = things to announce */
unsigned int mcastttl = 1; /* Multicast TTL */
char * ucsapaddr = NULL; /* Address for SAP Unicast announcements */
time_t startupts = 0; /* Startup-Timestamp */

#else /* not STANDALONE */

#include "sap.h"
#include "getstream.h"
#include "dvbrc.h"

#endif /* STANDALONE */

static int init_sap_timer(void );
static int sapsock;

/* For reference:
 * Global SAP address:                               224.2.127.254
 * Organization-local SAP address                    239.195.255.255
 * Local (smallest non-link-local scope) SAP address 239.255.255.255
 * Link-local SAP address                            224.0.0.255
 */

#define SAP_ADDR	"224.2.127.254"			/* Scope Global */
#define SAP_PORT	9875				/* As per RFC 2974 */
#define SAP_MAX_SIZE	1024				/* As per RFC 2974 */

#define SAP_ADDRTYPE_V4		0
#define SAP_ADDRTYPE_V6		1

#define SAP_MSGTYPE_ANNOUNCE	0
#define SAP_MSGTYPE_DELETE	1

#define SAP_VERSION		1

static uint8_t	sappkt[SAP_MAX_SIZE];

static void send_sap(int fd, short event, void *arg) {
	struct stream *		s;
	unsigned char *		sp;
	uint32_t		addr;
	int isvideo;

	s = streams;
	addr = inet_addr(SAP_ADDR);
	while (s) {
		if ((s->nuccs >= 0) &&
		    (ucsapaddr == NULL)) { /* Do not announce unicast streams */
			s = s->next;
			continue;
		}
		
#ifdef STANDALONE
		isvideo = s->isvideo;
#else
		if (s->cid < 0) { /* Do not announce special pids */
			s = s->next;
			continue;
		}
		isvideo = (s->pidcounters[PID_VIDEO] != 0);
#endif
		
		memset(&sappkt, 0, SAP_MAX_SIZE);
		/* First fill the 8 SAP header bytes */
		/* 0: Bitfield with version and various other info */
		sappkt[0] = (SAP_VERSION << 5)
		          + (SAP_ADDRTYPE_V4 << 4)
		          + (SAP_MSGTYPE_ANNOUNCE << 2);
		/* 1: authlen */
		sappkt[1] = 0x0;
		/* 2+3: msgidhash - sort of an unique ID */
		sappkt[2] = (s->channel->sid >> 8) & 0xff;
		sappkt[3] = (s->channel->sid >> 0) & 0xff;
		/* 4-7 (or 4-19 for IPv6): Origin. Together with the
		 * msgidhash this has to be unique. This should be our IP? */
		sappkt[4] = (addr >>  0) & 0xff;
		sappkt[5] = (addr >>  8) & 0xff;
		sappkt[6] = (addr >> 16) & 0xff;
		sappkt[7] = (addr >> 24) & 0xff;

		sp = &sappkt[8];
		sp += sprintf(sp, "v=0\r\n");
		sp += sprintf(sp, "o=- %x 1 IN IP4 %s:%d\r\n", s->channel->sid, s->ipaddr, s->port);
		sp += sprintf(sp, "s=%s %s\r\n", (isvideo) ? "TV" : "Radio", s->channel->name);
#if 0
		/* Announce a permanent stream. permanent usually means
		 * "more than 6 months" according to the RFC */
		sp += sprintf(sp, "t=0 0\r\n");
#else
		/* Announce a stream that times out in 5 minutes. Note that
		 * the timestamps are NTP timestamps, hence the -2208988800. */
		sp += sprintf(sp, "t=%lu %lu\r\n", startupts - 2208988800LU,
		                                   time(NULL) + 300 - 2208988800LU);
#endif
		/* RFC 1890 RTP Audio/Video Profile - MP2T - Mpeg 2 Transport Stream */
		if (s->nuccs >= 0) { /* Unicast */
			sp += sprintf(sp, "m=%s %d http 33\r\n", (isvideo) ? "video" : "audio", s->port);
#ifdef SUPPORTIPV6
			sp += sprintf(sp, "c=IN IP6 %s/%d\r\n", ucsapaddr, s->ttl);
#else
			sp += sprintf(sp, "c=IN IP4 %s/%d\r\n", ucsapaddr, s->ttl);
#endif
		} else {
			sp += sprintf(sp, "m=%s %d udp 33\r\n", (isvideo) ? "video" : "audio", s->port);
#ifdef SUPPORTIPV6
			sp += sprintf(sp, "c=IN IP6 %s/%d\r\n", s->ipaddr, s->ttl);
#else
			sp += sprintf(sp, "c=IN IP4 %s/%d\r\n", s->ipaddr, s->ttl);
#endif
		}
		sp += sprintf(sp, "a=tool:sapannouncer\r\n");
		sp += sprintf(sp, "a=type:broadcast\r\n");

		send(sapsock, sappkt, sp-sappkt, MSG_DONTWAIT);

		s = s->next;
	}

	init_sap_timer();
}


static int init_sap_timer(void ) {
	static struct event	sapevent;
	static struct timeval	tv;

	tv.tv_usec=0;
	tv.tv_sec=1;

	evtimer_set(&sapevent, send_sap, &sapevent);
	evtimer_add(&sapevent, &tv);

	return 1;
}

static int init_sap_sock(void ) {
	struct sockaddr_in	sin;
	unsigned int		ttl;
	struct ip_mreq		mreq;

	sapsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sapsock == -1) {
		perror("Failed to open sap socket");
		return 0;
	}

	sin.sin_family=AF_INET;
	inet_aton(SAP_ADDR, &sin.sin_addr);
	sin.sin_port=htons(SAP_PORT);

	if (connect(sapsock, (struct sockaddr *) &sin, sizeof(struct sockaddr_in)) != 0) {
		perror("sap socket connect");
		return 0;
	}

	ttl = mcastttl;
	setsockopt(sapsock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));

	/* Join the group */
	memcpy (&mreq.imr_multiaddr, &(sin.sin_addr), sizeof(struct in_addr));
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	if (setsockopt(sapsock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void*) &mreq, sizeof(mreq)) != 0) {
		perror("sap mcast join");
		return 0;
	}
	
	return 1;
}


int init_sap(void ) {
	init_sap_sock();
	init_sap_timer();

	return 1;
}

#ifdef STANDALONE

void showhelptxt(char * progname) {
	printf("Syntax: %s [-t ttl] [-h] [V:|A:]Name (unicast|multicast)addr1 ... \n", progname);
	printf("-t ttl         sets TTL for the SAP multicast packets (def. %u)\n", mcastttl);
}

void parsestreamparams(char * name, char * addr) {
	struct stream * tmpstr;
	char * spl;
	uint8_t tmps2[500];

	tmpstr = streams;
	streams = calloc(1, sizeof(struct stream));
	streams->next = tmpstr;
	streams->channel = calloc(1, sizeof(struct channel));
	streams->isvideo = 1;
	if (((name[0] == 'A') || (name[0] == 'V'))
	 && (name[1] == ':')) {
		if (name[0] == 'A') {
			streams->isvideo = 0;
		}
		name += 2;
	}
	streams->channel->name = strdup(name);
	streams->ttl = mcastttl;
	if (strncmp(addr, "udp://", 6) == 0) {
		addr += 6;
		streams->nuccs = -1;
	} else if (strncmp(addr, "http://", 7) == 0) {
		addr += 7;
		streams->nuccs = 0;
	} else {
		fprintf(stderr, "Unknown stream type in '%s'\n", addr);
		fprintf(stderr, "Only udp:// and http:// are supported.\n");
		exit(1);
	}
	if (strlen(addr) >= 500) {
		fprintf(stderr, "Error: URL too long (1) in '%s'\n", addr);
		exit(1);
	}
	strcpy(tmps2, addr);
	spl = strrchr(tmps2, ':');
	if (spl == NULL) {
		fprintf(stderr, "Failed to parse URL '%s'", addr);
		exit(1);
	}
	*spl = '\0';
	spl++;
	streams->port = strtoul(spl, NULL, 0);
	if (strlen(tmps2) >= sizeof(streams->ipaddr)) {
		fprintf(stderr, "Error: URL too long (2) in '%s'\n", addr);
		exit(1);
	}
	strcpy(streams->ipaddr, tmps2);
}

int main(int argc, char ** argv) {
	char	ch;
	int	argpos;
	
	startupts = time(NULL);
	event_init();
	
	while ((ch = getopt(argc, argv, "ht:")) != -1) {
		switch (ch) {
			case 'h':
				showhelptxt(argv[0]);
				exit(-1);
				break;
			case 't':
				mcastttl = strtoul(optarg, NULL, 10);
				break;
			default:
				fprintf(stderr, "Unknown command line arg\n");
				fprintf(stderr, "Try %s -h for help\n", argv[0]);
				exit(-1);
				break;
		}
	}
	argpos = optind;
	while (argpos < argc) {
		if ((argc - argpos) < 2) {
			fprintf(stderr, "ERROR: Non-Option arguments always go as pairs: First the name that is\n");
			fprintf(stderr, " to be announced, then the address that is to be announced.\n");
			fprintf(stderr, "In your arguments, '%s' has no corresponding address...\n", argv[argpos]);
			exit(-1);
		}
		parsestreamparams(argv[argpos], argv[argpos+1]);
		argpos += 2;
	}
	
	if (!streams) {
		fprintf(stderr, "ERROR: Need at least one thing to stream\n");
		exit(-1);
	}
	
	init_sap();
	
	/* Pigs can fly */
	event_dispatch();
	
	return 0;
}
#endif
