/* $Id: streamforwarder.c,v 1.17 2010/06/17 10:45:48 simimeie Exp $
 *
 * This little tool allows you to forward streams, e.g. between v6 and v4,
 * or unicast and multicast, ...
 */

#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>

#include <event.h>

#include "log.h"
#include "ttyhandlers.h"

/*
 * Defines
 */

#define MAX_MCAST_PAYLOAD	(1500-40)
#define TS_PACKET_SIZE		188
/* default readsize: as many full TS packets as we can put into
 * one multicast packet. */
#define USEFULREADSIZE ((MAX_MCAST_PAYLOAD / TS_PACKET_SIZE) * TS_PACKET_SIZE)

#define MAXSENDOVERFLOWS	500	/* After this many overflows the connection gets dropped */

#define HTTP_OK_REPLY "HTTP/1.0 200 OK\r\n"\
                      "Content-type: application/octet-stream\r\n"\
                      "\r\n"
#define HTTP_REQUEST_PART1 "GET "
#define HTTP_REQUEST_PART2 " HTTP/1.0\r\n"\
                           "\r\n"

/* Stream Types */
#define ST_HTTP		0
#define ST_MCAST	1
#define ST_FILE		2

#ifndef FILE_MAX
#define FILE_MAX 4096
#endif

/*
 * Types
 */
 
#ifdef SUPPORTIPV6
#define GS_AF	AF_INET6
#define GS_PF	PF_INET6
#define GS_SINA	sin6_addr
#define GS_SINP	sin6_port
#define GS_SINF sin6_family
	typedef struct ipv6_mreq gs_mreq;
	typedef struct sockaddr_in6 gs_sockaddr_struct;
#else /* ! IPV6 */
#define GS_AF	AF_INET
#define GS_PF	PF_INET
#define GS_SINA	sin_addr
#define GS_SINP	sin_port
#define GS_SINF sin_family
	typedef struct ip_mreq gs_mreq;
	typedef struct sockaddr_in gs_sockaddr_struct;
#endif

struct unicast_client {
	struct unicast_client *	next;
	int			sockfd;
	struct event		ev;
	unsigned long		sendoverflows;
	uint64_t		rxbytes;
	uint64_t		txbytes;
	char			hadrequest;
	gs_sockaddr_struct	remaddr;
};

struct streamresource {
	int			isopen;	/* boolean value - source open or not */
	int			fd;
	struct event		ev;
	int			type;
	char			fname[FILE_MAX]; /* Filename for ST_FILE, or textual IP/host else */
	uint8_t			ipaddr[128];
	uint16_t		port;	/* Port for ST_HTTP or ST_MCAST */
	signed char		httpdone; /* Gets set to 1 when we're through with the HTTP crap */
	int			nuccs;	/* Number of UniCast Clients for ST_HTTP */
	struct unicast_client *	ucclients; /* Clients for ST_HTTP */
};

struct stream {
	struct stream *		next;
	struct streamresource	src;
	struct streamresource	dst;
	int			pbufvalid;
	uint8_t 		pbuf[MAX_MCAST_PAYLOAD];
	time_t			lastdatarec;
	uint64_t		rxbytes;
	uint64_t		txbytes;
	unsigned int		reconnectdelay;
	time_t			lastconnect;
};

/*
 * Global Variables
 */

unsigned int	sndbuf_size = 65536 * 8; /* 512 KB */
int		readpersist = 0; /* 0 if we close the source when there are no listeners. */
struct stream *	streams = NULL;
unsigned int	maxuccs = 10; /* Maximum number of unicast clients */
int		mcastttl = 1;
unsigned long	upstreamtimeout = 10;
static struct event	toevent; /* Timeout Event Handler */
time_t		startupts = 0; /* Startup-Timestamp */
unsigned long	totconattempts = 0; /* Connection attempts (only unicast) */
unsigned long	totconsuccesss = 0; /* Successful connections */
const char * stdesc[] = { "HTTP", "MCAST", "FILE", "SEGFAULT" };
unsigned int	maxreconnectdelay = 120; /* Maximum reconnect delay */

/*
 * Functions
 */

void openstreamresource(struct streamresource * sr, int issrc, struct stream * parent);
static void timeoutcheck_init(void);

void sigpipehandler(int bla) { /* Dummyhandler for catching the event */
	return; /* It does nothing, but without it our program would die
		 * on receiving a SIGPIPE */
}

void init_sigpipehandler(void) {
	struct sigaction sia;
	sia.sa_handler = sigpipehandler;
	sigemptyset(&sia.sa_mask); /* If we don't do this, we're likely */
	sia.sa_flags = 0;          /* to die from 'broken pipe'! */
	sigaction(SIGPIPE, &sia, NULL);
}

void drop_tcp_connection(struct stream * str, int fd) {
	struct unicast_client * uc = str->dst.ucclients;
	struct unicast_client * prev = NULL;
	
	while (uc) {
		if (uc->sockfd == fd) { /* That's the one. */
			event_del(&uc->ev);
			close(uc->sockfd);
			if (prev == NULL) {
				str->dst.ucclients = uc->next;
			} else {
				prev->next = uc->next;
			}
			free(uc);
			str->dst.nuccs--;
			if ((str->dst.nuccs == 0) && (readpersist == 0)) {
				putlog(1, "[fd %05d]: Closing upstream fd, no more clients", str->src.fd);
				event_del(&str->src.ev);
				close(str->src.fd);
				str->src.isopen = 0;
				str->pbufvalid = 0;
			}
			return;
		}
		prev = uc;
		uc = uc->next;
	}
}

void dropupstreamforstream(struct stream * str) {
	event_del(&str->src.ev);
	close(str->src.fd);
	str->src.isopen = 0;
	str->pbufvalid = 0;
	str->reconnectdelay = (str->reconnectdelay + 1) * 2;
	if (str->reconnectdelay > maxreconnectdelay) {
		str->reconnectdelay = maxreconnectdelay;
	}
	/* Also drop all clients for that stream if HTTP */
	if (str->dst.type == ST_HTTP) {
		while (str->dst.ucclients) {
			struct unicast_client * uc = str->dst.ucclients;
			str->dst.ucclients = uc->next;
			putlog(1, "[fd %05d]: Warning: dropping client fd due to upstream fd %d error", uc->sockfd, str->src.fd);
			event_del(&uc->ev);
			close(uc->sockfd);
			free(uc);
		}
		str->dst.nuccs = 0;
	}
}

void setfdnonblocking(int fd) {
	int flags;
	flags = fcntl(fd, F_GETFL, 0);
	flags |= O_NONBLOCK|O_NDELAY;
	if (fcntl(fd, F_SETFL, flags) < 0) {
		perror("nonblocking fcntl");
		exit(1);
	}
}

void showhelptxt(char * progname) {
	printf("Syntax: %s [options] (pipe|stream)addr1 (unicast|multicast|file)addr1 ... \n", progname);
	printf("-d level       Sets the log level. Higher level gives more output (defaults to %d)\n", console_loglevel);
	printf("-b sndbuf      How much bufferspace to request from the OS (SO_SNDBUF; default %u)\n", sndbuf_size);
	printf("-l limit       Limit for number of connections PER STREAM (!). (def. %u)\n", maxuccs);
	printf("               Used for unicast only (there are no connections on mcast)\n");
	printf("-t ttl         sets TTL for multicast packets (def. %u)\n", mcastttl);
}

void parseurl(char * what, struct streamresource * sr) {
	char * spl;
	uint8_t tmpstr[FILE_MAX];
	
	memset(sr, 0, sizeof(struct streamresource));
	sr->type = ST_HTTP;
	if        (strncmp(what, "file://", 7) == 0) {
		sr->type = ST_FILE;
		what += 7;
	} else if (strncmp(what, "udp://", 6) == 0) {
		sr->type = ST_MCAST;
		what += 6;
	} else if (strncmp(what, "http://", 7) == 0) {
		sr->type = ST_HTTP;
		what += 7;
	}
	strncpy(&sr->fname[0], what, FILE_MAX-1); /* Terminating 0 from memset above */
	if (sr->type == ST_FILE) {
		/* We're finished already. */
		return;
	}
	/* For mcast or http: split into hostname/ip and port, and resolve */
	strcpy(tmpstr, &sr->fname[0]);
	spl = strrchr(tmpstr, ':');
	if (spl == NULL) {
		fprintf(stderr, "Failed to parse URL '%s'", what);
		exit(1);
	}
	*spl = '\0';
	spl++;
#ifdef SUPPORTIPV6
	if (!inet_pton(AF_INET6, tmpstr, &sr->ipaddr[0])) {
		/* No V6 - try v4 */
		char v4addrinv6style[sizeof(tmpstr)+10];
		sprintf(v4addrinv6style, "::ffff:%s", tmpstr);
		if (!inet_pton(AF_INET6, v4addrinv6style, &sr->ipaddr[0])) {
			fprintf(stderr, "ERROR: IP address '%s' is not valid.\n", tmpstr);
			exit(1);
		}
	}
#else /* ! IPV6 */
	if (!inet_aton(tmpstr, &sr->ipaddr[0])) {
		fprintf(stderr, "ERROR: IP address '%s' is not valid.\n", tmpstr);
		exit(1);
	}
#endif
	sr->port = strtoul(spl, NULL, 0);
}

static inline void send_packet(struct stream * str, uint8_t * buf) {
	int len;
	int tosend = str->pbufvalid;
	if        (str->dst.type == ST_MCAST) {
		len = send(str->dst.fd, str->pbuf, tosend, MSG_DONTWAIT);
		if (len != tosend) {
			putlog(0, "send didn't send all... %d/%d", len, tosend);
		}
		if (len > 0) { str->txbytes += len; }
	} else if (str->dst.type == ST_HTTP) {
		/* Send to all registered clients */
		struct unicast_client * uc = str->dst.ucclients;
		while (uc) {
			if (!uc->hadrequest) {
				uc = uc->next;
				continue;
			}
			len = write(uc->sockfd, str->pbuf, tosend);
			if (len != tosend) {
				putlog(9, "[fd %05d]: failed send for the %luth time... %d/%d",
					uc->sockfd, uc->sendoverflows, len, tosend);
				uc->sendoverflows++;
				if (uc->sendoverflows > MAXSENDOVERFLOWS) {
					struct unicast_client * olduc = uc;
					uc = uc->next;
					putlog(1, "[fd %05d]: Too many overflows, dropping connection.",
						  olduc->sockfd);
					drop_tcp_connection(str, olduc->sockfd);
					continue;
				}
			}
			if (len > 0) {
				str->txbytes += len;
				uc->txbytes += len;
			}
			uc = uc->next;
		}
	} else if (str->dst.type == ST_FILE) {
		len = write(str->dst.fd, str->pbuf, tosend);
		if (len != tosend) {
			putlog(0, "write didn't write all... %d/%d", len, tosend);
		}
		if (len > 0) { str->txbytes += len; }
	} else {
		/* WTF? */
	}
}

void data_incoming(int fd, short event, void * up) {
	int len;
	struct stream * str;
	int readsize = USEFULREADSIZE;
	
	str = (struct stream *)up;
	if (str->src.fd != fd) {
		putlog(0, "!!! %s called with nonmatching data: fd = %d, str->src.fd = %d !!!",
		       __FUNCTION__, fd, str->src.fd);
		return;
	}
	if ((event & EV_WRITE) == EV_WRITE) {
		if ((str->src.httpdone < 0) && (str->src.type == ST_HTTP)) {
			char * filepartofurl;
			int cres;
			int cressize = sizeof(cres);
			/* Check if connection attempt was successful */
			if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &cres, &cressize)) {
				putlog(0, "!!! %s: getsockopt failed. !!!", __FUNCTION__);
				dropupstreamforstream(str);
				return;
			}
			if (sizeof(cres) != cressize) {
				putlog(0, "!!! %s: connection result is not an int. !!!", __FUNCTION__);
				dropupstreamforstream(str);
				return;
			}
			if (cres != 0) {
				putlog(1, "[fd %05d]: Upstream connect failed: %s", fd, strerror(errno));
				dropupstreamforstream(str);
				return;
			}
			putlog(1, "[fd %05d]: Upstream connect succeeded.", fd);
			/* Register BEFORE the request - else we miss the reply if the server is fast! DOH! */
			event_set(&str->src.ev, str->src.fd, EV_READ | EV_PERSIST, data_incoming, str);
			event_add(&str->src.ev, NULL);
			/* Send HTTP request */
			write(fd, HTTP_REQUEST_PART1, strlen(HTTP_REQUEST_PART1));
			filepartofurl = index(str->src.fname, '/');
			if (filepartofurl == NULL) {
				filepartofurl = "/";
			}
			write(fd, filepartofurl, strlen(filepartofurl));
			write(fd, HTTP_REQUEST_PART2, strlen(HTTP_REQUEST_PART2));
			str->src.httpdone = 0;
			return;
		} else {
			putlog(1, "[fd %05d]: WARNING: unexpected write ready event on source socket: %d", fd, event);
			return;
		}
	}
	if ((event & ~(EV_WRITE | EV_READ)) != 0) {
		putlog(1, "[fd %05d]: WARNING: unknown event on source socket: %d", fd, event);
		return;
	}
	/* For multicast, read as much as possible, because if we read only
	 * part of a packet, the rest of it gets dropped, unless it came
	 * from localhost, then we get what we expect, the rest of the packet
	 * on the next read. How intuitive. */
	if (str->src.type == ST_MCAST) {
		readsize = MAX_MCAST_PAYLOAD;
	}
	/* Make sure not to read too much */
	if ((str->pbufvalid + readsize) > MAX_MCAST_PAYLOAD) {
		readsize = MAX_MCAST_PAYLOAD - str->pbufvalid;
		if ((str->src.type != ST_MCAST) &&
		    ((str->pbufvalid + readsize) < USEFULREADSIZE)) {
			/* Just fill the packet */
			readsize = USEFULREADSIZE - str->pbufvalid;
		}
	}
	assert(readsize > 0);
	len = read(fd, &str->pbuf[str->pbufvalid], readsize);
	if (len > 0) {
		str->lastdatarec = time(NULL);
		str->rxbytes += len;
		if ((str->src.type == ST_HTTP) && (str->src.httpdone == 0) && (len >= 4)) {
			int j;
			for (j = 0; j < (len-3); j++) {
				if ((str->pbuf[j+0] == '\r')
				 && (str->pbuf[j+1] == '\n')
				 && (str->pbuf[j+2] == '\r')
				 && (str->pbuf[j+3] == '\n')) {
					putlog(1, "[fd %05d]: HTTP done on upstream", str->src.fd);
					str->src.httpdone = 1;
					len -= (j + 4);
					memmove(&str->pbuf[0], &str->pbuf[j+4], len);
					break;
				}
			}
			if (str->src.httpdone == 0) {
				for (j = 0; j < (len-1); j++) {
					if ((str->pbuf[j+0] == '\n')
					 && (str->pbuf[j+1] == '\n')) {
						putlog(1, "[fd %05d]: HTTP done on upstream (despite buggy httpd)", str->src.fd);
						str->src.httpdone = 1;
						len -= (j + 2);
						memmove(&str->pbuf[0], &str->pbuf[j+2], len);
						break;
					}
				}
			}
			if (str->src.httpdone == 0) {
				len = 0;
			}
			if (len < 0) {
				putlog(1, "[fd %05d]: Warning: len < 0?! %d", str->src.fd, len);
				len = 0;
			}
		}
		str->pbufvalid += len;
		if ((str->pbufvalid + readsize) >= MAX_MCAST_PAYLOAD) {
			/* Another packet won't fit, so send now */
			send_packet(str, str->pbuf);
			str->pbufvalid = 0;
		}
	} else {
		/* Not sure how to handle this? Probably best to close */
		putlog(1, "[fd %05d]: Warning: Closed upstream fd due to error", fd);
		dropupstreamforstream(str);
	}
}

void work_client_connection(int fd, short event, void * up) {
	struct stream * str = (struct stream *)up;
	char buf[1024];
	
	if (event & EV_READ) {
		int rdres;
		rdres = read(fd, buf, 1023);
		if (rdres <= 0) {
			putlog(1, "[fd %05d]: Read error, dropping connection", fd);
			drop_tcp_connection(str, fd);
			return;
		} else {
			struct unicast_client * uc = str->dst.ucclients;
			while (uc) {
				if (uc->sockfd == fd) {
					if (!uc->hadrequest) {
						write(uc->sockfd,
						      HTTP_OK_REPLY,
						      strlen(HTTP_OK_REPLY));
						uc->hadrequest = 1;
					}
					uc->rxbytes += rdres;
				}
				uc = uc->next;
			}
		}
	}
}

void new_tcp_connection(int fd, short event, void * up) {
	struct stream *	s = (struct stream *)up;
	gs_sockaddr_struct remaddr;
	unsigned int sl = sizeof(gs_sockaddr_struct);
	int newsock;
	struct unicast_client *	uc;
	int sopt;
	
	newsock = accept(fd, (struct sockaddr *)&remaddr, &sl);
	if (newsock < 0) {
		perror("(ignoring error) ACCEPT");
		return;
	}
	totconattempts++;
	openstreamresource(&s->src, 1, s);
#ifdef SUPPORTIPV6
	if (1) {
		char translbuf[INET6_ADDRSTRLEN];
		if (inet_ntop(AF_INET6, &remaddr.sin6_addr,
		              translbuf, sizeof(translbuf)) == NULL ) {
			strcpy(translbuf, "!ERROR!");
		}
		putlog(1, "[fd %05d]: New connection from '%s' to stream from fd %d",
		          newsock, translbuf, s->src.fd);
	}
#else /* ! IPV6 */
	putlog(1, "[fd %05d]: New connection from %s to stream from fd %d",
	          newsock, inet_ntoa(remaddr.sin_addr), s->src.fd);
#endif /* ! IPV6 */

	if (s->src.isopen == 0) {
		putlog(1, "[fd %05d]: Dropping new client, upstream is unavailable.",
		          newsock);
		close(newsock);
		return;
	}
	if (s->dst.nuccs >= maxuccs) {
		putlog(1, "[fd %05d]: Too many connections already, dropping the new one.",
		          newsock);
		close(newsock);
		return;
	}
	totconsuccesss++;
	setfdnonblocking(newsock);
	sopt = sndbuf_size;
	if (setsockopt(newsock, SOL_SOCKET, SO_SNDBUF, (char *)&sopt, sizeof(sopt))) {
		perror("setsockopt SNDBUF");
	}
	
	uc = calloc(1, sizeof(struct unicast_client));
	uc->sockfd = newsock;
	uc->remaddr = remaddr;
	uc->next = s->dst.ucclients;
	s->dst.ucclients = uc;
	s->dst.nuccs++;
	
	event_set(&uc->ev, uc->sockfd, EV_READ | EV_PERSIST, work_client_connection, s);
	event_add(&uc->ev, NULL);
}

void openstreamresource(struct streamresource * sr, int issrc, struct stream * parent) {
	if (sr->isopen) {
		return;
	}
	if (issrc) {
		parent->lastconnect = time(NULL);
	}
	if (sr->type == ST_FILE) {
		/* Open the file (which should be a pipe, otherwise it wouldn't make much sense) */
		sr->fd = open(sr->fname, ((issrc) ? O_RDONLY : O_WRONLY) | O_NONBLOCK);
		if (sr->fd < 0) {
			putlog(1, "stream resource %s: %s (%s) ...Failed: open: %s",
			          stdesc[sr->type], sr->fname,
			          ((issrc) ? "input" : "output"), strerror(errno));
			return;
		}
		putlog(1, "[fd %05d]: Opening stream resource: %s: %s (%s)",
		          sr->fd, stdesc[sr->type],
		          sr->fname, ((issrc) ? "input" : "output"));
		if (issrc) {
			event_set(&sr->ev, sr->fd, EV_READ | EV_PERSIST, data_incoming, parent);
			event_add(&sr->ev, NULL);
		}
	} else {
		/* MCast or http - open socket */
		gs_sockaddr_struct	sin;
		int sopt;
		
		sin.GS_SINF = GS_AF;
		memcpy((char *)&sin.GS_SINA, &sr->ipaddr[0], sizeof(sin.GS_SINA));
		sin.GS_SINP = htons(sr->port);
		
		if (sr->type == ST_MCAST) { /* Multicast address */
			sr->fd = socket(GS_PF, SOCK_DGRAM, IPPROTO_UDP);
			if (sr->fd < 0) {
				putlog(1, "stream resource %s: %s (%s) ...Failed: socket(UDP): %s",
				          stdesc[sr->type], sr->fname,
				          ((issrc) ? "input" : "output"), strerror(errno));
				return;
			}
			putlog(1, "[fd %05d]: Opening stream resource: %s: %s (%s)",
			          sr->fd, stdesc[sr->type],
			          sr->fname, ((issrc) ? "input" : "output"));
			if (issrc) { /* Data source */
				gs_mreq joinreq;
				sopt = 1;
				if (setsockopt(sr->fd, SOL_SOCKET, SO_REUSEADDR, &sopt, sizeof(sopt)) < 0) {
					putlog(1, "...Warning: nonfatal error on setsockopt: %s", strerror(errno));
				}
				if (bind(sr->fd, (struct sockaddr *) &sin, sizeof(sin))) {
					putlog(1, "...Warning: nonfatal error on bind: %s", strerror(errno));
				}
				/* Join multicast group */
#ifdef SUPPORTIPV6
				/* FIXME Untested! */
				joinreq.ipv6mr_interface = 0; /* Default interface */
				memcpy((char *)&joinreq.ipv6mr_multiaddr, (char *)&sr->ipaddr[0], sizeof(joinreq.ipv6mr_multiaddr));
				if (setsockopt(sr->fd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, (char *)&joinreq, sizeof(joinreq))) {
#else /* ! IPV6 */
				joinreq.imr_interface.s_addr = INADDR_ANY; /* Default interface */
				memcpy((char *)&joinreq.imr_multiaddr.s_addr, (char *)&sr->ipaddr[0], sizeof(uint32_t));
				if (setsockopt(sr->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&joinreq, sizeof(joinreq))) {
#endif /* ! IPV6 */
					putlog(1, "...Failed: setsockopt IP_ADD_MEMBERSHIP: %s", strerror(errno));
				}
				setfdnonblocking(sr->fd);
				event_set(&sr->ev, sr->fd, EV_READ | EV_PERSIST, data_incoming, parent);
				event_add(&sr->ev, NULL);
			} else { /* Data destination */
				sopt = sndbuf_size;
				if (setsockopt(sr->fd, SOL_SOCKET, SO_SNDBUF, (char *)&sopt, sizeof(sopt))) {
					putlog(1, "...Failed: setsockopt SNDBUF: %s", strerror(errno));
				}
				if (connect(sr->fd, (struct sockaddr *)&sin, sizeof(sin))) {
					putlog(1, "...Failed: connect: %s", strerror(errno));
					return;
				}
				/* Set socket TTL */
#ifdef SUPPORTIPV6
				if (setsockopt(sr->fd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, (char *)&mcastttl, sizeof(mcastttl)) != 0) {
#else /* ! IPV6 */
				if (setsockopt(sr->fd, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&mcastttl, sizeof(mcastttl)) != 0) {
#endif /* ! IPV6 */
					putlog(1, "...Warning: nonfatal error on setsockopt MULTICAST_TTL: %s", strerror(errno));
					/* This is not fatal, so don't exit. */
				}
			}
		} else { /* Unicast address */
			sr->fd = socket(GS_PF, SOCK_STREAM, IPPROTO_TCP);
			if (sr->fd < 0) {
				putlog(1, "stream resource %s: %s (%s) ...Failed: socket(TCP): %s",
				          stdesc[sr->type], sr->fname,
				          ((issrc) ? "input" : "output"), strerror(errno));
				return;
			}
			putlog(1, "[fd %05d]: Opening stream resource: %s: %s (%s)",
			          sr->fd, stdesc[sr->type],
			          sr->fname, ((issrc) ? "input" : "output"));
			if (issrc) { /* Data source */
				/* Connect */
				setfdnonblocking(sr->fd);
				/* Since we're nonblocking, we register for EV_WRITE, that's the event we will get
				 * once the connect request finishes (successful or not) */
				event_set(&sr->ev, sr->fd, EV_WRITE, data_incoming, parent);
				event_add(&sr->ev, NULL);
				if (connect(sr->fd, (struct sockaddr *)&sin, sizeof(gs_sockaddr_struct))) {
					if (errno != EINPROGRESS) {
						putlog(1, "...Failed: connect: %s", strerror(errno));
						return;
					}
				}
				sr->httpdone = -1;
			} else { /* Data destination */
				setfdnonblocking(sr->fd);
				sopt = 1;
				if (setsockopt(sr->fd, SOL_SOCKET, SO_REUSEADDR, (char *)&sopt, sizeof(sopt))) {
					putlog(1, "...Warning: nonfatal error on setsockopt REUSEADDR: %s", strerror(errno));
				}
				if (bind(sr->fd, (struct sockaddr *) &sin, sizeof(sin))) {
					putlog(1, "...Failed: bind: %s", strerror(errno));
					return;
				}
				if (listen(sr->fd, 10)) {
					putlog(1, "...Failed: listen: %s", strerror(errno));
					return;
				}
				sr->nuccs = 0;
				event_set(&sr->ev, sr->fd, EV_READ | EV_PERSIST, new_tcp_connection, parent);
				event_add(&sr->ev, NULL);
			}
		}
	}
	sr->isopen = 1;
	parent->lastdatarec = time(NULL);
}

void parsestreamparams(char * what, char * whereto) {
	struct stream * tmpstr;

	tmpstr = streams;
	streams = calloc(1, sizeof(struct stream));
	streams->next = tmpstr;
	parseurl(what,    &streams->src);
	parseurl(whereto, &streams->dst);
	if ((readpersist) || (streams->dst.type != ST_HTTP)) {
		openstreamresource(&streams->src, 1, streams);
	}
	openstreamresource(&streams->dst, 0, streams);
}

unsigned char * mytabcompleter(unsigned char * curinp) {
	if (strcmp(curinp, "q") == 0) { return("quit"); }
	if (strcmp(curinp, "qu") == 0) { return("quit"); }
	if (strcmp(curinp, "qui") == 0) { return("quit"); }
	return NULL;
}

static uint8_t * printtimediff(time_t d) {
	static uint8_t opbuf[100];
	uint8_t secs, mins, hours, days;
	
	secs = d % 60;
	d /= 60;
	mins = d % 60;
	d /= 60;
	hours = d % 24;
	d /= 24;
	days = d;
	sprintf(opbuf, "%3d+%02d:%02d:%02d", days, hours, mins, secs);
	return &opbuf[0];
}

static uint8_t * printtrafcounter(unsigned long long bc) {
	static uint8_t opbuf[100];
	uint8_t tmpunit = ' ';
	
	if (bc >= 100000ULL) {
		tmpunit = 'K';
		bc /= 1000ULL;
	}
	if (bc >= 100000ULL) {
		tmpunit = 'M';
		bc /= 1000ULL;
	}
	if (bc >= 100000ULL) {
		tmpunit = 'G';
		bc /= 1000ULL;
	}
	sprintf(opbuf, "%9llu%c", bc, tmpunit);
	return &opbuf[0];
}

static void doshowversion(void) {
	printf("Streamforwarder Version %s%s Compiled on %s %s\n",
	       "0.2", "+beta1", __DATE__, __TIME__);
}

static void doshowstreams(void) {
	struct stream * s = streams;
	printf("%-20s isopen? %-20s isopen? %-5s %10s %10s\n", "src", "dst", "nuccs", "RX", "TX");
	while (s) {
		printf("%-5s %-20s %1d ", stdesc[s->src.type], s->src.fname, s->src.isopen);
		printf("%-5s %-20s %1d ", stdesc[s->dst.type], s->dst.fname, s->dst.isopen);
		printf("%5d ", s->dst.nuccs);
		printf("%s ", printtrafcounter(s->txbytes));
		printf("%s ", printtrafcounter(s->rxbytes));
		printf("\n");
		s = s->next;
	}
}

static void doshowclients(void) {
	struct stream * s = streams;
	unsigned long totclients = 0;
	printf("%-5s %-39s %-5s %10s %10s %s\n", "FD", "IP", "OvFl.", "TX", "RX", "Stream");
	while (s) {
		if (s->dst.type == ST_HTTP) {
			struct unicast_client * uc = s->dst.ucclients;
			while (uc) {
#ifdef SUPPORTIPV6
				char translbuf[INET6_ADDRSTRLEN];
				if (inet_ntop(AF_INET6, &uc->remaddr.sin6_addr,
				              translbuf, sizeof(translbuf)) == NULL ) {
					strcpy(translbuf, "!ERROR!");
				}
#endif /* IPV6 */
				printf("%05d %-39s %5lu ",
				       uc->sockfd,
#ifdef SUPPORTIPV6
				       translbuf,
#else /* ! IPv6 */
				       inet_ntoa(uc->remaddr.sin_addr),
#endif /* ! IPV6 */
				       uc->sendoverflows);
				printf("%s ", printtrafcounter(uc->txbytes));
				printf("%s ", printtrafcounter(uc->rxbytes));
				printf("%s%s %s\n",
				       (uc->hadrequest ? "" : "!"),
				       stdesc[s->src.type], s->src.fname);
				totclients++;
				
				uc = uc->next;
			}
		}
		s = s->next;
	}
	printf("--- End of list - %lu unicast client%s total\n",
	       totclients, ((totclients == 1) ? "" : "s"));
}

static void doshowstatus(void) {
	struct stream * s = streams;
	unsigned long nforwards = 0;
	unsigned long mcastsrc = 0;
	unsigned long ucastsrc = 0;
	unsigned long filesrc = 0;
	unsigned long mcastdst = 0;
	unsigned long ucastdst = 0;
	unsigned long filedst = 0;
	unsigned long nucclients = 0;
	unsigned long long tottxbytes = 0;
	unsigned long long totrxbytes = 0;
	
	printf("General Server Status:\n");
	while (s) {
		nforwards++;
		if (s->src.type == ST_HTTP) { ucastsrc++; }
		if (s->src.type == ST_MCAST) { mcastsrc++; }
		if (s->src.type == ST_FILE) { filesrc++; }
		if (s->dst.type == ST_HTTP) {
			ucastdst++;
			nucclients += s->dst.nuccs;
		}
		if (s->dst.type == ST_MCAST) { mcastdst++; }
		if (s->dst.type == ST_FILE) { filedst++; }
		tottxbytes += s->txbytes;
		totrxbytes += s->rxbytes;
		s = s->next;
	}
	printf("%lu forwards configured.\n", nforwards);
	printf("  Sources:      %3lu file, %3lu multicast, %3lu unicast\n", filesrc, mcastsrc, ucastsrc);
	printf("  Destinations: %3lu file, %3lu multicast, %3lu unicast\n", filedst, mcastdst, ucastdst);
	printf("%lu unicast client%s connected\n", nucclients, ((nucclients == 1) ? "" : "s"));
	printf("%20llu Bytes (%s) received from the network (avg: %5.1lf MBit)\n",
	       (unsigned long long)totrxbytes,
	       printtrafcounter(totrxbytes),
	       (double)(totrxbytes/(((double)(time(NULL) - startupts)) * (1000.0 / 8.0) * 1000.0)));
	printf("%20llu Bytes (%s) sent to the network (avg: %5.1lf MBit)\n",
	       (unsigned long long)tottxbytes,
	       printtrafcounter(tottxbytes),
	       (double)(tottxbytes/(((double)(time(NULL) - startupts)) * (1000.0 / 8.0) * 1000.0)));
	printf("%lu successful connections, %lu connection attempts (%5.1lf%%)\n",
	       totconsuccesss, totconattempts,
	       (double)((double)(totconsuccesss * 100) / (double)totconattempts));
	printf("Uptime: %s\n", printtimediff(time(NULL) - startupts));
}

void handleterminput(int fd, short event, void * up) {
	char buf[10];
	
	if (event & EV_READ) {
		int rdres;
		rdres = read(fd, buf, 1);
		if (rdres != 1) {
			return;
		}
		if (addtoinputbuf(buf[0])) {
			unsigned char * kbdcmd;
			
			kbdcmd = getinputline();
			putlog(0, "[Console command]: %s", kbdcmd);
			jmptocol1();
			cleartoeol();
			if        (strcmp(kbdcmd, "status") == 0) {
				doshowstatus();
				clearinputbuf();
			} else if (strcmp(kbdcmd, "showclients") == 0) {
				doshowclients();
				clearinputbuf();
			} else if (strcmp(kbdcmd, "showstreams") == 0) {
				doshowstreams();
				clearinputbuf();
			} else if ((strcmp(kbdcmd, "showversion") == 0)
			        || (strcmp(kbdcmd, "version") == 0)) {
				doshowversion();
				clearinputbuf();
			} else if (strcmp(kbdcmd, "quit") == 0) {
				clearinputbuf();
				printf("Goodbye cruel world.\n");
				putlog(0, "--- QUITTING through quit command from console ---");
				jmptocol1();
				cleartoeol();
				exit(0);
			} else if (strcmp(kbdcmd, "help") == 0) {
				printf("Available commands:\n");
				printf("help           shows this text\n");
				printf("quit           quits\n");
				printf("showclients    lists current (unicast) clients\n");
				printf("showstreams    lists streams\n");
				printf("status         shows general status\n");
				clearinputbuf();
			} else {
				printf("Unknown command.\n");
			}
			redrawinputline();
		}
	}
}

static void timeoutchecker(int fd, short event, void *arg) {
	/* Check all streams for timeouts */
	struct stream * s = streams;
	while (s) {
		if ((s->src.type == ST_HTTP) && (s->src.isopen)) {
			if ((time(NULL) - s->lastdatarec) > upstreamtimeout) {
				putlog(1, "[fd %05d]: Warning: Closed upstream fd due to timeout", s->src.fd);
				dropupstreamforstream(s);
			}
		}
		/* Also check for upstreams we should attempt to restore */
		if (s->src.isopen == 0) { /* Stream not open? */
			if ((readpersist) || (s->dst.type != ST_HTTP)) { /* Stream of a type we have to reconnect? */
				if ((time(NULL) - s->lastconnect) > s->reconnectdelay) {
					putlog(1, "[fd N/A  ]: Reconnecting: lastconnect %u  time %u  reconnectdelay %u", s->lastconnect, time(NULL), s->reconnectdelay);
					openstreamresource(&s->src, 1, s);
				}
			}
		} else {
			/* While the stream is successfully open, slowly decrease
			 * the reconnectdelay back to 0 again. */
			if ((s->reconnectdelay > 0) && ((time(NULL) - s->lastdatarec) < 2)) {
				s->reconnectdelay--;
			}
		}
		s = s->next;
	}
	/* Reinit timer */
	timeoutcheck_init();
}

static void timeoutcheck_init(void) {
	struct timeval tv;
	
	tv.tv_usec = 0;
	tv.tv_sec = 1;
	evtimer_set(&toevent, timeoutchecker, &toevent);
	evtimer_add(&toevent, &tv);
}

int main(int argc, char ** argv) {
	char	ch;
	int	argpos;
	static struct event	termevent;
	
	doshowversion();
	startupts = time(NULL);
	
	init_sigpipehandler();
	event_init();
	
	while ((ch = getopt(argc, argv, "b:d:l:hpt:")) != -1) {
		switch (ch) {
			case 'b':
				sndbuf_size = strtoul(optarg, NULL, 10);
				break;
			case 'd':
				console_loglevel = strtol(optarg, NULL, 10);
				break;
			case 'h':
				showhelptxt(argv[0]);
				exit(-1);
				break;
			case 'l':
				maxuccs = strtoul(optarg, NULL, 10);
				break;
			case 'p':
				readpersist = 1;
				break;
			case 't':
				mcastttl = strtoul(optarg, NULL, 10);
				break;
			default:
				fprintf(stderr, "Unknown command line arg\n");
				fprintf(stderr, "Try %s -h for help\n", argv[0]);
				exit(-1);
				break;
		}
	}
	argpos = optind;
	while (argpos < argc) {
		if ((argc - argpos) < 2) {
			fprintf(stderr, "ERROR: Non-Option arguments always go as pairs: First what to forward, then where to forward it to.\n");
			fprintf(stderr, "In your arguments, '%s' has no place to be forwarded to...\n", argv[argpos]);
			exit(-1);
		}
		parsestreamparams(argv[argpos], argv[argpos+1]);
		argpos += 2;
	}
	
	if (!streams) {
		fprintf(stderr, "ERROR: Need at least one thing to stream\n");
		exit(-1);
	}

	/* Initialize terminal */
	putlog(5, "Terminal init...");
	if (setupstdin()) {
		event_set(&termevent, STDIN_FILENO, EV_READ | EV_PERSIST, handleterminput, NULL);
		event_add(&termevent, NULL);
		registertabcompletioncallback(mytabcompleter);
	}
	
	putlog(5, "Timeoutcheck init...");
	timeoutcheck_init();
	
	/* Pigs can fly */
	putlog(0, "running.");
	event_dispatch();

	return 0;
}
